﻿using SimpleJSON;
using System;
using System.Linq;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using UnityEngine.EventSystems;
using uFileBrowser;
using Battlehub.RTCommon;
using System.Linq.Expressions;
using UnityThreading;
using System.Text;
using System.Text.RegularExpressions;
using MVR.FileManagementSecure;

namespace VAM_Decal_Maker
{
    public class Decal_Maker : MVRScript
    {
        const string pluginName = "DecalMaker";
        const string pluginVersion = "Beta 1";
        //person script is attatched too
        Atom _parentAtom; //VAM parent
        Atom _shaderAtom; //Obj holding shader
        bool _SetupFinished = false;
        int _queuedBakes = 0;

        public Shader _customDecalShader { get; private set; }
        public Shader _customSpecGlossShader { get; private set; }
        public Shader _customNormShader { get; private set; }
        public Shader _customPackedNormShader { get; private set; }
        public Shader _cutoutShader { get; private set; }
        
        public Dictionary<string, int[]> _TextureIndex { get; private set; }


        public Texture2D _clearTex { get; private set; }
        public DAZCharacterTextureControl _dazCharactertextureControl { get; private set; }
        public DAZMergedSkinV2 _dazSkin { get; private set; }
  

        //UI
        JSONStorableBool _toggleNippleCutout = new JSONStorableBool("Nipple Cutouts ON", true);
        JSONStorableBool _toggleGenitalCutout = new JSONStorableBool("Genital Cutouts ON", true);


        UIDynamicButton _resetButton;
        UIDynamicButton _saveButton;
        UIDynamicButton _clearAllButton;
        UIDynamicButton _prefabButton;
        UIDynamicButton _savePresetButton;
        UIDynamicButton _loadPresetButton;

        UIDynamicToggle _toggleNipple;
        UIDynamicToggle _toggleGenital;

        HeaderPanel _headerPanel;

        bool _Debug = false;

        public override void Init()
        {
            try
            {
                if (containingAtom.mainController == null)
                {
                    SuperController.LogError("Please add this plugin to a PERSON atom.");
                    return;
                }
                _parentAtom = containingAtom;
                _parentAtom.mainController.interactableInPlayMode = true;
                //load custom asset bundle at get the customshader
                StartCoroutine(CreateAssetAtom("Chokpahi-Asset", "Chokpahi-Asset"));

                _clearAllButton = CreateButton("Clear Everything");
                _prefabButton = CreateButton("USE Prefab", true);

				PrefabPanel prefabPanel = new PrefabPanel(this);

                _resetButton = CreateButton("Reset Texture");
                _saveButton = CreateButton("Apply Texture", true);

                _loadPresetButton = CreateButton("Load Preset");
                _savePresetButton = CreateButton("Save Preset", true);

                _toggleNipple = CreateToggle(_toggleNippleCutout);
                _toggleGenital = CreateToggle(_toggleGenitalCutout, true);

                RegisterBool(_toggleNippleCutout);
                RegisterBool(_toggleGenitalCutout);
                
                //assign listners
                _clearAllButton.button.onClick.AddListener(ResetAll);

                _prefabButton.button.onClick.AddListener(prefabPanel.OnClick);

                _saveButton.button.onClick.AddListener(() =>
                {
                    TempManager.DestoryAll();
                    _headerPanel.Bake();
                });

                string presetsPath = "Custom/Scripts/Chokaphi/VAM_Decal_Maker/Presets";
                if (!MVR.FileManagementSecure.FileManagerSecure.DirectoryExists(presetsPath))
                {
                    MVR.FileManagementSecure.FileManagerSecure.CreateDirectory(presetsPath);
                }
                if (!MVR.FileManagementSecure.FileManagerSecure.DirectoryExists(presetsPath + "/PreFabs"))
                {
                    MVR.FileManagementSecure.FileManagerSecure.CreateDirectory(presetsPath + "/PreFabs");
                }
              
                _savePresetButton.button.onClick.AddListener(() =>
                {
                    SuperController.singleton.fileBrowserUI.defaultPath = presetsPath;
                    SuperController.singleton.fileBrowserUI.showFiles = true;
                    SuperController.singleton.fileBrowserUI.showDirs = true;
                    SuperController.singleton.fileBrowserUI.SetTextEntry(true);
                    SuperController.singleton.fileBrowserUI.hideExtension = true;
                    SuperController.singleton.fileBrowserUI.Show(SavePresetCallback);
                });

                _loadPresetButton.button.onClick.AddListener(() =>
                {
                    string presets = "Custom/Scripts/Chokaphi/VAM_Decal_Maker/Presets";

                    SuperController.singleton.fileBrowserUI.defaultPath = presets;
                    SuperController.singleton.fileBrowserUI.SetTitle("Select Load Decal Preset");
                    SuperController.singleton.fileBrowserUI.showFiles = true;
                    SuperController.singleton.fileBrowserUI.showDirs = true;
                    SuperController.singleton.fileBrowserUI.SetTextEntry(false);
                    SuperController.singleton.fileBrowserUI.hideExtension = true;
                    SuperController.singleton.fileBrowserUI.Show(PresetLoadCallBack);
                });
               
                byte[] tmppng = MVR.FileManagementSecure.FileManagerSecure.ReadAllBytes(GetPackagePath(this) + "Custom/Scripts/Chokaphi/VAM_Decal_Maker/Cutout/Clear.png");
                _clearTex = new Texture2D(1, 1, TextureFormat.RGBA32, false);
                _clearTex.LoadImage(tmppng);

                JSONStorableAction resetAllAction = new JSONStorableAction("ClearAll", ResetAll);
                RegisterAction(resetAllAction);

            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        void LogError(string err)
        {
            if (_Debug)
            {
                SuperController.LogError(err);
            }
        }

        void DebugSave(Texture2D tempTex, string name)
        {
            //string path = "Custom/Scripts/Chokaphi/OUT/" + name;
            //byte[] bytes = tempTex.EncodeToPNG();

            //MVR.FileManagementSecure.FileManagerSecure.WriteAllBytes(path, bytes);
        }
        //MacGruber Discord 03/26/2020
        //Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
        public static string GetPluginPath(MVRScript self)
        {
            string id = self.name.Substring(0, self.name.IndexOf('_'));
            string filename = self.manager.GetJSON()["plugins"][id].Value;
            return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
        }

        // Get path prefix of the package that contains our plugin.
        public string GetPackagePath()
        {
            return GetPackagePath(this);
        }
        public static string GetPackagePath(MVRScript self)
        {
            string filename = GetPluginPath(self);
            int idx = filename.IndexOf(":/");
            if (idx >= 0)
                return filename.Substring(0, idx + 2);
            else
                return string.Empty;
        }
        //By: VAMDeluxe Discord
        public List<string> GetFilesAtPathRecursive(string path, string pattern)
        {
            List<string> combined = new List<string>();
            string[] files = MVR.FileManagementSecure.FileManagerSecure.GetFiles(path, pattern);
            string[] directories = MVR.FileManagementSecure.FileManagerSecure.GetDirectories(path);
            

            files.ToList().ForEach(file =>
            {
                combined.Add(file);
            });

            directories.ToList().ForEach(directory =>
            {
                combined.AddRange(GetFilesAtPathRecursive(directory, pattern));
            });

            return combined;
        }
        //from C# PATH
        char[] sep = new char[] { '/', '\\' };
        public string GetFileName(string path)
        {
            int num = path.LastIndexOfAny(sep);
            if (num >= 0)
            {
                return path.Substring(num + 1);
            }
            return path;
        }

        //"project-canyon12/05/2018 DISCORD"
        //"I'm going on how I understand Async/Await works in normal C#, but I think the principle is the same"
        //tweaked for my use case
        delegate void AtomCreatedEvent(Atom atom);
        IEnumerator CreateAssetAtom(string uniqueAtomName, string nameOfGameObjectToFind)
        {
            yield return new WaitForSeconds(0.5f);

            Atom createdAtom = GetAtomById(uniqueAtomName);
            if (createdAtom == null)
            {
                yield return SuperController.singleton.AddAtomByType("CustomUnityAsset", uniqueAtomName, true);

                createdAtom = GetAtomById(uniqueAtomName);
            }

            JSONStorable storable = createdAtom.GetStorableByID("asset");
            storable.SetUrlParamValue("assetUrl", GetPackagePath(this) + @"Custom\Assets\Chokaphi\chokpahi-decal.assetbundle");
            storable.SetStringChooserParamValue("assetName", "Assets/Chokpahi-Decal.unity");

            yield return new WaitWhile(() => GameObject.Find(nameOfGameObjectToFind) == null);

            LogError("Asset loaded");
            _shaderAtom = createdAtom;

            yield return new WaitForSeconds(0.5f);
            Shader[] renderers = Resources.FindObjectsOfTypeAll<Shader>();
            foreach (Shader shader in renderers)
            {
                if (shader.name == "Custom/Unlit-Decal2")
                {
                    LogError("ShaderFind " + Shader.Find(shader.name) + " Custom Shader is " + shader.name);
                    _customDecalShader = shader;
                }
                if (shader.name == "Custom/Unlit-Cutout2")
                {
                    LogError("ShaderFind " + Shader.Find(shader.name) + " Cutout Shader is " + shader.name);
                    _cutoutShader = shader;
                }
                if (shader.name == "Custom/Unlit-Specular")
                {
                    LogError("ShaderFind " + Shader.Find(shader.name) + " Spec-Gloss Shader is " + shader.name);
                    _customSpecGlossShader = shader;
                }
                if (shader.name == "Custom/Unlit-NormalMap")
                {
                    LogError("ShaderFind " + Shader.Find(shader.name) + " Normal Shader is " + shader.name);
                    _customNormShader = shader;
                }
                if (shader.name == "Custom/Unlit-PackedNormalMap")
                {
                    LogError("ShaderFind " + Shader.Find(shader.name) + " PackedNorm Shader is " + shader.name);
                    _customPackedNormShader = shader;
                }
            }

        }

        // Update is called with each rendered frame by Unity
        void Update()
        {
            try
            {
                //need to delay creation on Sceen reloads till VAM has finished loading
                if (_parentAtom && _customDecalShader && _dazCharactertextureControl == null)
                {
                    _dazCharactertextureControl = _parentAtom.GetComponentInChildren<DAZCharacterTextureControl>();
                    if (_dazCharactertextureControl != null)
                    {
                        _dazSkin = _dazCharactertextureControl.GetComponent<DAZMergedSkinV2>();
                        if (_dazSkin != null)
                        {
                            if (_SetupFinished)
                            {
                                _headerPanel.UpdateSelectionPanels();
                                StartCoroutine(ApplyDecalsThreaded(1f));
                            }
                            else
                            {
                                Setup();
                            }
                        }
                    }
                }

            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        ////A public event for listeners to subscribe to
        //public event EventHandler SomethingHappened;

        

        //public void OnDestroy()
        //{
        //    //SuperController.LogError("DESTORY");
        //    ////Fire the event - notifying all subscribers
        //    //if (SomethingHappened != null)
        //    //    SomethingHappened(this, null);

        //}

        //public override void PreRestore()
        //{
        //    SuperController.LogError("************************************************PRE restore");
        //}

        //public override void PostRestore()
        //{
        //    SuperController.LogError("--------------------------------------------POST restore");
        //}

        //this is called on SKIN changes from the Update Event
        public void Setup()
        {
            try
            {
                _TextureIndex = new Dictionary<string, int[]>()
                {
                    {"torso" , _dazCharactertextureControl.torsoMaterialNums },
                    { "face" , _dazCharactertextureControl.faceMaterialNums },
                    { "limbs", _dazCharactertextureControl.limbMaterialNums },
                    { "genitals", _dazCharactertextureControl.genitalMaterialNums },
                };

                //create Header Panel
                _headerPanel = new HeaderPanel(this);

                _resetButton.button.onClick.AddListener(_headerPanel.ResetTextures);

                _headerPanel.UpdateSelectionPanels();
    
                //UIDynamicButton NewPanelButton = CreateButton(" LAST LEFT ELEMENT");              
                //UIDynamicButton UnusedButton2 = CreateButton("LAST RIGHT ELEMENT", true);

                //this is hacky but sets the UI to look nice on start.
                _prefabButton.button.onClick.Invoke();
                _headerPanel.SelectionPanelSpec.TorsoPanel.Button.onClick.Invoke();
                _headerPanel.SelectionPanelGloss.TorsoPanel.Button.onClick.Invoke();
                _headerPanel.SelectionPanelNorm.TorsoPanel.Button.onClick.Invoke();


                _headerPanel.decalButton.button.onClick.Invoke();
                _headerPanel.SelectionPanelDecal.TorsoPanel.Button.onClick.Invoke();


                _SetupFinished = true;
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        public void UpdateSelectionPanels(SelectionPanel selectionPanel)
        {
            LogError("UpdateSelectionPanels " + selectionPanel.PanelName);
            foreach (var panel in selectionPanel.Selections)
            {
                int index = _TextureIndex[panel.Key].FirstOrDefault();
                string slot = panel.Value.MaterialSlot;

                Texture2D temp = null;
                if (slot == "_DecalTex")
                {
                    temp = (Texture2D)_dazSkin.GPUmaterials[index].GetTexture("_MainTex");
                }
                else if (slot == "_BumpMap")
                {   //convert any packed normals to rgb format
                    Texture2D bump = (Texture2D)_dazSkin.GPUmaterials[index].GetTexture(slot);

                    if (bump && bump.format == TextureFormat.DXT5)
                {
                        LogError("Bump map format is dxt5 " + panel.Key + " " + slot);
                        temp = GpuCombine(_dazSkin.GPUmaterials[index].GetTexture(slot), _customPackedNormShader, slot, Color.white, null, true);
                }
                    else
                {
                        temp = bump;
                    }
                }
                else
                {
                    temp = (Texture2D)_dazSkin.GPUmaterials[index].GetTexture(slot);
                }

                panel.Value.image.material.SetTexture("_MainTex" , temp);
                panel.Value.Original = _dazSkin.GPUmaterials[index].GetTexture(slot);
            }

        }

        private void ResetAll()
        {
            _headerPanel.ResetTextures();

            foreach (var v in _headerPanel.selectionPanels)
            {
                foreach (var selPanel in    v.Value.Selections)
                {
                    int count = selPanel.Value.DecalPanels.Count;
                    for (int i = 0; i < count; i++)
                    {
                        selPanel.Value.RemoveDecalPanel();
                    }
                }
            }
        }

        public void ResetTextures(ImagePanel panel)
        {
            foreach (int num in _TextureIndex[panel.TextureSlot])
            {
                _dazSkin.GPUmaterials[num].SetTexture(panel.MaterialSlot, panel.Original);
            }
            TempManager.DestoryAll();
        }
        
        public void RemoveTempDecalPanels()
        {
            _headerPanel.RemoveTempDecalPanels();
        }

        public void ResetTempDecalPanelTable()
        {
            _headerPanel.ResetTempDecalPanelTable();
        }

        //used when restoring from json or profile.
        //make sure all images are loaded before applying
        private IEnumerator ApplyDecalsThreaded(float delay = 0.5f)
        {
            yield return new WaitForSeconds(delay);

            foreach (var panels in _headerPanel.selectionPanels)
            {
                foreach (var selPanel in panels.Value.Selections)
                {
                    if (selPanel.Value.DecalPanels.Count > 0)
                    {
                        foreach (DecalPanel d in selPanel.Value.DecalPanels)
                        {
                            yield return new WaitWhile(() => d.ImagePanel.ImageLoading == true);
                        }
                    }
                }
            }
            TempManager.DestoryAll();
            _headerPanel.Bake();
        }

        public void Bake(List<DecalPanel> DecalPanels, string TextureSlot, string MaterialSlot)
        {
            if (!_customDecalShader || !_cutoutShader || !_customNormShader || !_customSpecGlossShader)
            {
                SuperController.LogError("DECALMAKER:CUSTOM SHADER NOT FOUND.");
                return;
            }
            _queuedBakes++;
            StartCoroutine(BakeThreaded(DecalPanels, TextureSlot, MaterialSlot));
        }
        private IEnumerator BakeThreaded(List<DecalPanel> DecalPanels, string TextureSlot, string MaterialSlot)
        {
            yield return new WaitForEndOfFrame();
            switch (MaterialSlot)
            {
                case "_DecalTex":
                    StartCoroutine(BakeDecal(DecalPanels, TextureSlot, MaterialSlot));
                    break;
                case "_SpecTex":
                case "_GlossTex":
                    StartCoroutine(BakeSpecular(DecalPanels, TextureSlot, MaterialSlot));
                    break;
                case "_BumpMap":
                    StartCoroutine(BakeNormal(DecalPanels, TextureSlot, MaterialSlot));
                    break;
            }

            yield return new WaitWhile(() => _queuedBakes > 0);
            TempManager.DestoryTempObjs();
        }

        private IEnumerator BakeDecal(List<DecalPanel> DecalPanels, string TextureSlot, string MaterialSlot)
        {
            int count = 0;
         
            Texture2D tempTex = TempManager.CopyTexture(_clearTex);
            foreach (DecalPanel d in DecalPanels)
            {
                if (d.ImagePanel.image.mainTexture != null)
                {
                    yield return new WaitForEndOfFrame();
                    tempTex = GpuCombine(tempTex, _customDecalShader, MaterialSlot, d.Color, d.ImagePanel.image.mainTexture);
                    DebugSave(tempTex, TextureSlot + "-" + MaterialSlot + " " + count + ".png");
                    count++;
                }
            }
            //apply cutouts
            if (TextureSlot == "torso" && _toggleNippleCutout.val && _cutoutShader != null)
            {
                yield return new WaitForEndOfFrame();
                byte[] png = MVR.FileManagementSecure.FileManagerSecure.ReadAllBytes(GetPackagePath(this) + "Custom/Scripts/Chokaphi/VAM_Decal_Maker/Cutout/" + _dazCharactertextureControl.uvSetName + ".png");
                Texture2D alphaTex = TempManager.TempTexture2D(1, 1, TextureFormat.RGBA32, false);
                alphaTex.LoadImage(png);
                tempTex = GpuClip(tempTex, alphaTex);
            }
            if (TextureSlot == "genitals" && _toggleGenitalCutout.val && _cutoutShader != null)
            {
                yield return new WaitForEndOfFrame();
                byte[] png = MVR.FileManagementSecure.FileManagerSecure.ReadAllBytes(GetPackagePath(this) + "Custom/Scripts/Chokaphi/VAM_Decal_Maker/Cutout/_FemaleGenitals.png");
                Texture2D alphaTex = TempManager.TempTexture2D(1, 1, TextureFormat.RGBA32, false);
                alphaTex.LoadImage(png);
                tempTex = GpuClip(tempTex, alphaTex);
            }

            foreach (int num in _TextureIndex[TextureSlot])
            {
                yield return new WaitForEndOfFrame();
                _dazSkin.GPUmaterials[num].SetTexture(MaterialSlot, tempTex);
            }

            TempManager.MakeSessionObj(tempTex);
            _queuedBakes--;
            yield break;
        }
        //handles Specular and Gloss linear textures
        private IEnumerator BakeSpecular(List<DecalPanel> DecalPanels, string TextureSlot, string MaterialSlot)
        {
            //Texture2D tempTex = (Texture2D)_dazSkin.GPUmaterials[_TextureIndex[TextureSlot][0]].GetTexture(MaterialSlot);
            Texture2D tempTex = null;
            int count = 0;
            foreach (DecalPanel d in DecalPanels)
            {
                yield return new WaitForEndOfFrame();
                if (d.ImagePanel.image.mainTexture != null)
                {
                   
                    if (tempTex == null)
                    {
                        tempTex = TempManager.CopyTexture((Texture2D)d.ImagePanel.image.mainTexture, true);
                    }
                    else
                    {
                        yield return new WaitForEndOfFrame();
                        tempTex = GpuCombine(tempTex, _customSpecGlossShader, MaterialSlot, d.Color, d.ImagePanel.image.mainTexture, true);
                    }
                    DebugSave(tempTex, TextureSlot + "-" + MaterialSlot + " " + count + ".png");
                    count++;
                }
            }
            foreach (int num in _TextureIndex[TextureSlot])
            {
                yield return new WaitForEndOfFrame();
                _dazSkin.GPUmaterials[num].SetTexture(MaterialSlot, tempTex);
            }

            TempManager.MakeSessionObj(tempTex);
            _queuedBakes--;
            yield break;
        }
   
        private IEnumerator BakeNormal(List<DecalPanel> DecalPanels, string TextureSlot, string MaterialSlot)
        {
            _dazSkin.GPUmaterials[_TextureIndex[TextureSlot][0]].EnableKeyword("_NORMALMAP");
           
            //LogError("queue amount = " + _queuedBakes);
            //Texture2D tempTex = (Texture2D)_dazSkin.GPUmaterials[_TextureIndex[TextureSlot][0]].GetTexture(MaterialSlot);
            Texture2D tempTex = null;
            int count = 0;
            foreach (DecalPanel d in DecalPanels)
            {
                yield return new WaitForEndOfFrame();
                if (d.ImagePanel.image.mainTexture != null)
                {
                    if (tempTex == null)
                    {
                        tempTex = TempManager.CopyTexture((Texture2D)d.ImagePanel.image.mainTexture, true);
                    }
                    else
                    { 
                        yield return new WaitForEndOfFrame();
                        tempTex = GpuCombine(tempTex, _customNormShader, MaterialSlot, d.Color, d.ImagePanel.image.mainTexture, true); 
                    }
                    DebugSave(tempTex, TextureSlot + "-" + MaterialSlot + " " + count + ".png");
                    count++;
                }
            }

            foreach (int num in _TextureIndex[TextureSlot])
            {
                _dazSkin.GPUmaterials[num].SetTexture(MaterialSlot, tempTex);
            }

            TempManager.MakeSessionObj(tempTex);
            _queuedBakes--;
            yield break;
        }

        public void PresetLoadCallBack(string path)
        {
            PresetLoad(path);
        }

        public void PresetLoad(string path, bool temp = false)
        {
            if (string.IsNullOrEmpty(path))
                return;

            JSONClass jc = this.LoadJSON(path) as JSONClass;
            PerformLoad(jc, temp);
        }

        public void SavePresetCallback(string path)
        {
            if (string.IsNullOrEmpty(path))
                return;

            JSONClass json = GetJSON();
            path = path.Replace(".dsgn.DecalMakerPreset", "");
            path = path + "." + DSGN + ".DecalMakerPreset.json";
            this.SaveJSON(json, path);
            SuperController.singleton.DoSaveScreenshot(path);
            SuperController.singleton.fileBrowserUI.ClearCacheImage(path);
        }

        //SAVE
        private string DSGN = "dsgn";
        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, true);
            jc["SaveVersion"] = "1";

            //save
            jc["Decal"] = _headerPanel.SelectionPanelDecal.Save();
            jc["Specular"] = _headerPanel.SelectionPanelSpec.Save();
            jc["Gloss"] = _headerPanel.SelectionPanelGloss.Save();
            jc["Normal"] = _headerPanel.SelectionPanelNorm.Save();


            jc["Nipple Cutouts ON"] = _toggleNippleCutout.val.ToString();
            jc["Genital Cutouts ON"] = _toggleGenitalCutout.val.ToString();

            return jc;
        }

        //LOAD
        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms);
            StartCoroutine(PerformLoadThreaded(jc));
        }
        //called once we have everything loaded. RestorefromJSON is too early
        private IEnumerator PerformLoadThreaded(JSONClass jc)
        {
            yield return new WaitUntil(() => _SetupFinished == true && _customDecalShader != null);
            PerformLoad(jc);
        }

        private void PerformLoad(JSONClass jc, bool temp = false)
        {
            if (jc == null)
                return;

            //convert to new file version if save is old
            try
            {
            if (jc["SaveVersion"] == null)
                jc = ConvertSaveToV1(jc);

            }
            catch
            {
                SuperController.LogError("Decal Maker: Unable to convert save");
                return;
            }
            // Load Save Data
            try
            {
            if (jc["Nipple Cutouts ON"] != null)
            {
                _toggleNippleCutout.SetVal(bool.Parse(jc["Nipple Cutouts ON"]));
            }

            if (jc["Genital Cutouts ON"] != null)
            {
                _toggleGenitalCutout.SetVal(bool.Parse(jc["Genital Cutouts ON"]));
            }

            foreach (var panel in _headerPanel.selectionPanels)
            {
                if (jc[panel.Key] != null)
                {
                    foreach (var selPanel in panel.Value.Selections)
                    {
                        foreach (JSONClass saveJSON in jc[panel.Key][selPanel.Key] as JSONArray)
                        {
                                DecalPanel d = panel.Value.AddDecalPanel(selPanel.Key, temp);
                            if (saveJSON["Alpha"] != null)
                            {
                                d.sliderDynamic.slider.value = (float.Parse(saveJSON["Alpha"]));
                            }
                            if (saveJSON["H"] != null)
                            {
                                float h = float.Parse(saveJSON["H"]);
                                float s = float.Parse(saveJSON["S"]);
                                float v = float.Parse(saveJSON["V"]);
                                d.jcolor.SetVal(h, s, v);
                            }
                            if (saveJSON["Path"] != null)
                            {
                                d.BrowserCallBack(saveJSON["Path"]);
                            }
                        }
                    }
                }
            }
           
            StartCoroutine(ApplyDecalsThreaded());
        }
            catch (Exception e)
            {
                SuperController.LogError("Decal Maker: Unable to Load save");
                SuperController.LogError(e.Message);
                return;
            }
        }
        //upgrades old saves to new format
        private JSONClass ConvertSaveToV1(JSONClass jc)
        {
            List<string> keys = new List<string>() { "torso", "genitals","face","limbs" };
            JSONClass dc = new JSONClass();
            foreach (string key in keys)
            {
                if (jc[key] != null)
                {
                    dc.Add(key, jc[key]);
                }
                jc.Remove(key);
            }
            jc["Decal"] = dc;
        
            return jc;
        }

        //VAM uses System Drawing see ImageLoaderThreaded Process()
        public Texture2D GetTextureFromSurfaceShader(Material mat, int width, int height)
        {
            //Create render texture:
            RenderTexture temp = new RenderTexture(width, height, 0, RenderTextureFormat.ARGB32, RenderTextureReadWrite.sRGB);
            temp.DiscardContents();
            //Create a Quad:
            GameObject quad = GameObject.CreatePrimitive(PrimitiveType.Quad);
            MeshRenderer rend = quad.GetComponent<MeshRenderer>();

            rend.material = mat;

            // Vector3 quadScale = quad.transform.localScale / (float)((Screen.height / 2.0) / Camera.main.orthographicSize);
            Vector3 quadScale = quad.transform.localScale;
            quad.transform.position = Vector3.forward;

            //Setup camera:
            GameObject camGO = new GameObject("CaptureCam");
            Camera cam = camGO.AddComponent<Camera>();
            cam.renderingPath = RenderingPath.Forward;
            cam.orthographic = true;
            cam.clearFlags = CameraClearFlags.SolidColor;
            cam.backgroundColor = new Color(1, 1, 1, 0);
            if (cam.rect.width < 1 || cam.rect.height < 1)
            {
                cam.rect = new Rect(cam.rect.x, cam.rect.y, 1, 1);
            }
            cam.orthographicSize = 0.5f;
            cam.rect = new Rect(0, 0, quadScale.x, quadScale.y);
            cam.aspect = quadScale.x / quadScale.y;
            cam.targetTexture = temp;
            cam.allowHDR = false;


            //Capture image and write to the render texture:
            cam.Render();
            temp = cam.targetTexture;

            //Apply changes:
            Texture2D newTex = new Texture2D(temp.width, temp.height, TextureFormat.RGBA32, true, false);
            RenderTexture.active = temp;
            newTex.ReadPixels(new Rect(0, 0, temp.width, temp.height), 0, 0);
            newTex.Apply();

            //Clean up:
            RenderTexture.active = null;
            temp.Release();
            Destroy(quad);
            Destroy(camGO);

            return newTex;
        }

        //Uses shader to clip part of one texture and applys it to another based on the alpha of a third texture.
        //Used to Apply "Clean" nipples and genital areas after decals have been applied.
        //This Alpha control will be Character UV specific. So Victoria has a diffrent nipple area than Olympia etc..
        public Texture2D GpuClip(Texture mainTex, Texture alphaTex)
        {
            //LogError("GPU Clip textures " + _bitsShader.name);

            int h = mainTex.width;
            int w = mainTex.height;

            RenderTexture tmp = RenderTexture.GetTemporary(w, h, 0, RenderTextureFormat.ARGB32, RenderTextureReadWrite.sRGB);

            // Backup the currently set RenderTexture
            RenderTexture previous = RenderTexture.active;

            Material material = TempManager.TempMaterial(_cutoutShader);
            material.mainTexture = mainTex;
            material.SetTexture("_Alpha", alphaTex);

            Graphics.Blit(mainTex, tmp, material);
            RenderTexture.active = tmp;

            Texture2D newTexture = TempManager.TempTexture2D(w, h, TextureFormat.RGBA32, false);
            newTexture.ReadPixels(new Rect(0, 0, w, h), 0, 0);
            newTexture.Apply();

            // Reset the active RenderTexture
            RenderTexture.active = previous;
            RenderTexture.ReleaseTemporary(tmp);

            return newTexture;
        }
        public Texture2D GpuCombine(Texture mainTex, Shader shader, string materialSlot, Color color, Texture overlayTex = null, bool linear = false)
        {
            if (mainTex == null)
            { 
                return null; 
            }

            int h = mainTex.width;
            int w = mainTex.height;

            RenderTextureReadWrite renderTexConversion = RenderTextureReadWrite.sRGB;
            if (linear)
            {
                renderTexConversion = RenderTextureReadWrite.Linear;
            }

             RenderTexture tmp = RenderTexture.GetTemporary(w, h, 0, RenderTextureFormat.ARGB32, renderTexConversion);
            //tmp.antiAliasing = 8;
            //Backup the currently set RenderTexture
            RenderTexture previous = RenderTexture.active;

            Material material = TempManager.TempMaterial(shader);
            material.mainTexture = mainTex;
                
            if (materialSlot == "_DecalTex")
            {
                material.SetVector("_DecalColor", color); 
            }
            if (overlayTex)
            {
                material.SetTexture(materialSlot, overlayTex);
            }

            Graphics.Blit(mainTex, tmp, material);
            RenderTexture.active = tmp;

            Texture2D newTexture = TempManager.TempTexture2D(w, h, TextureFormat.RGBA32, false, linear);
            newTexture.ReadPixels(new Rect(0, 0, w, h), 0, 0, false);
            newTexture.Apply();

            // Reset the active RenderTexture
            RenderTexture.active = previous;
            RenderTexture.ReleaseTemporary(tmp);

            return newTexture;
        }
        //old method
        public Texture2D GpuCombine(Texture mainTex, Texture decalTex, Color color)
        {
            int h = mainTex.width;
            int w = mainTex.height;

            RenderTexture tmp = RenderTexture.GetTemporary(w, h, 0, RenderTextureFormat.ARGB32, RenderTextureReadWrite.sRGB);
            //tmp.antiAliasing = 8;
            // Backup the currently set RenderTexture
            RenderTexture previous = RenderTexture.active;

            Material material = TempManager.TempMaterial(_customDecalShader);
            material.mainTexture = mainTex;
            material.SetTexture("_DecalTex", decalTex);
            material.SetVector("_DecalColor", color);

            Graphics.Blit(mainTex, tmp, material);
            RenderTexture.active = tmp;

            Texture2D newTexture = TempManager.TempTexture2D(w, h, TextureFormat.RGBA32, false, false);
            newTexture.ReadPixels(new Rect(0, 0, w, h), 0, 0, false);
            newTexture.Apply();

            // Reset the active RenderTexture
            RenderTexture.active = previous;
            RenderTexture.ReleaseTemporary(tmp);

            return newTexture;
        }
        //From https://support.unity3d.com/hc/en-us/articles/206486626-How-can-I-get-pixels-from-unreadable-textures-
        public Texture2D GetUnreadableTexture(Texture2D texture, bool linear = false)
        {
            RenderTextureReadWrite renderTexConversion = RenderTextureReadWrite.sRGB;
            if (linear)
            {
                renderTexConversion = RenderTextureReadWrite.Linear;
            }
            //LogError("Get Unreadable texture");
            RenderTexture tmp = RenderTexture.GetTemporary(texture.width, texture.height, 0, RenderTextureFormat.ARGB32, renderTexConversion);

            // Blit the pixels on texture to the RenderTexture
            Graphics.Blit(texture, tmp);
            // Backup the currently set RenderTexture
            RenderTexture previous = RenderTexture.active;
            // Set the current RenderTexture to the temporary one we created
            RenderTexture.active = tmp;

            Texture2D newTexture = TempManager.TempTexture2D(texture.width, texture.height, TextureFormat.RGBA32, false, linear);

            newTexture.ReadPixels(new Rect(0, 0, tmp.width, tmp.height), 0, 0);
            newTexture.Apply();

            // Reset the active RenderTexture
            RenderTexture.active = previous;
            // Release the temporary RenderTexture
            RenderTexture.ReleaseTemporary(tmp);
            //LogError("Finish Getting Unreadable texture");
            return newTexture;
        }

    }

    public class PrefabPanel
    {
        public GameObject gameObject = new GameObject("PresetPanel");
        public ImagePanel ImagePanel { get; private set; }

        private Dictionary<string, List<PrefabCollection>> prefabFiles = new Dictionary<string, List<PrefabCollection>>();

        private UIDynamic spacerLeft;
        private UIDynamic spacerRight;
        private Decal_Maker _DM;

        private string selectedType;
        private string selectedItem;

        List<string> author = new List<string>();
        List<string> prefabTypes = new List<string>();
        List<string> prefabNames = new List<string>();

        JSONStorableStringChooser jsc;
        JSONStorableStringChooser jsc2;
        public PrefabPanel(Decal_Maker DM)
        {
            _DM = DM;
            //create a UI Spacer to Hold our Panel
            spacerLeft = DM.CreateSpacer(false);
            spacerLeft.height = 550;
            spacerRight = DM.CreateSpacer(true);
            spacerRight.height = 550;

            ImagePanel = new ImagePanel();
            ImagePanel.gameObject.transform.SetParent(spacerRight.transform, false);
            RectTransform imageRect = ImagePanel.image.GetComponent<RectTransform>();
            imageRect.sizeDelta = new Vector2(500, 500);

            GridLayoutGroup glg = spacerLeft.gameObject.AddComponent<GridLayoutGroup>();
            glg.cellSize = new Vector2(500, 100);

            glg.constraint = GridLayoutGroup.Constraint.FixedColumnCount;
            glg.constraintCount = 1;
            glg.padding = new RectOffset(10, 10, 10, 10);
            glg.spacing = new Vector2(10, 0);


            Transform popupPrefabType = GameObject.Instantiate<Transform>(_DM.manager.configurableScrollablePopupPrefab);
            popupPrefabType.SetParent(spacerLeft.gameObject.transform, false);
            popupPrefabType.localPosition = Vector3.zero;

            UIDynamicPopup dynamicPopupType = popupPrefabType.GetComponent<UIDynamicPopup>();
            dynamicPopupType.name = "CLASS";
            //popupToJSONStorableStringChooser.Add(PrefabType, jsc);
            jsc = new JSONStorableStringChooser("name", prefabTypes, "", "Choose Type");
            jsc.popup = dynamicPopupType.popup;

            
            Transform popupPrefabItem = GameObject.Instantiate<Transform>(_DM.manager.configurableScrollablePopupPrefab);
            popupPrefabItem.SetParent(spacerLeft.gameObject.transform, false);
            popupPrefabItem.localPosition = Vector3.zero;

            UIDynamicPopup dynamicPopupItem = popupPrefabItem.GetComponent<UIDynamicPopup>();
            dynamicPopupItem.name = "ITEM";
            dynamicPopupItem.popup.showSlider = false;
            
            RefreshPrefabCollection();
            jsc2 = new JSONStorableStringChooser("name", prefabNames, "", "Choose Prefab");
            jsc2.popup = dynamicPopupItem.popup;

            dynamicPopupType.popup.onValueChangeUnityEvent.AddListener(TypeValueChanges);
            dynamicPopupItem.popup.onValueChangeUnityEvent.AddListener(ItemValueChanged);

            ImagePanel.Button.onClick.AddListener(() =>
            {
                _DM.ResetTempDecalPanelTable();
            });

            GameObject imageObject = new GameObject("text background");
            imageObject.transform.SetParent(ImagePanel.gameObject.transform, false);
            Image ima = imageObject.AddComponent<Image>();
            ima.color =new Color(.8f,.8f,.8f);
            RectTransform imRT = imageObject.GetComponent<RectTransform>();
            imageObject.transform.localPosition += new Vector3(0,-5f, 0);
            imRT.anchorMin = Vector2.zero;
            imRT.anchorMax = new Vector2(1, 0.1f);
            imRT.sizeDelta = Vector2.zero;


            //Button Text
            GameObject textObject = new GameObject("button text");
            textObject.transform.SetParent(imageObject.gameObject.transform, false);
            //used to prevent text object from blocking clicks
            CanvasGroup cg = textObject.AddComponent<CanvasGroup>();
            cg.blocksRaycasts = false;

            Text panelText = textObject.AddComponent<Text>();
            panelText.alignment = TextAnchor.MiddleCenter;
            panelText.color = Color.black;
            panelText.font = Resources.GetBuiltinResource<Font>("Arial.ttf");
            panelText.fontSize = 30;
            panelText.fontStyle = FontStyle.Bold;
            panelText.text = "Make current Prefab Permanent";

            RectTransform ptRT = panelText.GetComponent<RectTransform>();
            ptRT.anchorMin = new Vector2(0, .02f);
            ptRT.anchorMax = Vector2.one;
            ptRT.sizeDelta = Vector2.zero;
 
        }

        private void TypeValueChanges(string value)
        {
            selectedType = value;
            RefreshPrefabCollection();
            List<PrefabCollection> a = prefabFiles[value];
            prefabNames = a.Select(x => x.ItemName).ToList();
            prefabNames.Insert(0, "");
            jsc2.SetVal("");
            ResetItem();
            jsc2.choices = prefabNames;
        }

        private void ItemValueChanged(string value)
        {
            selectedItem = value;
            _DM.RemoveTempDecalPanels();
            if (value == "")
            {
                ResetItem();
            }
            else
            {
                PrefabCollection c = prefabFiles[selectedType].Where(x => x.ItemName == value).FirstOrDefault();
                //reset decals added from other prefabs
                if (c.ImagePath != null)
                    ImagePanel.LoadPNG(c.ImagePath, ImagePanel);

                _DM.PresetLoad(c.Path, true);
            }
        }

        private void ResetItem()
        {
            _DM.RemoveTempDecalPanels();
            ImagePanel.ClearTexture();
            _DM.PresetLoad(_DM.GetPackagePath() +"Custom/Scripts/Chokaphi/VAM_Decal_Maker/Icons/EMPTY.dsgn.DecalMakerPreset.json", true);
        }

        public void RefreshPrefabCollection()
        {
            string path = "Custom/Scripts/Chokaphi/VAM_Decal_Maker/Presets/Prefabs";
            if (!MVR.FileManagementSecure.FileManagerSecure.DirectoryExists(path))
                return;

            List<string> files = _DM.GetFilesAtPathRecursive(path, null);
            
            prefabFiles = new Dictionary<string, List<PrefabCollection>>();
            foreach (string file in files)
            {
                PrefabCollection pc = new PrefabCollection(file);
                if (pc.ItemType != null)
                {
                    if (prefabFiles.ContainsKey(pc.ItemType))
                    {
                        prefabFiles[pc.ItemType].Add(pc);
                    }
                    else
                    {
                        prefabFiles.Add(pc.ItemType, new List<PrefabCollection>() { pc });
                    }
                }
            }
            prefabTypes = prefabFiles.Keys.ToList();
            jsc.choices = prefabTypes;
        }

        public void OnClick()
        {
            bool state = !spacerLeft.gameObject.activeSelf;
            spacerLeft.gameObject.SetActive(state);
            spacerRight.gameObject.SetActive(state);
        }
    }

    public class DecalPanel
    {
        public GameObject gameObject = new GameObject("DecalPanel");
        public ImagePanel ImagePanel { get; private set; }

        public MyButton Up { get; private set; }
        public MyButton Down { get; private set; }
        public MyButton Open { get; private set; }
        public MyButton Close { get; private set; }

        public Color Color { get { return ImagePanel.image.color; } }

        public JSONStorableColor jcolor { get; private set; }

        public UIDynamicSlider sliderDynamic { get; private set; }

        public Action<string> LastDir;

        public Text panelText { get; private set; }

        Decal_Maker _DM;
        public DecalPanel(Transform anchor, float width, float height, Color? color, Decal_Maker DM, bool isNormalMap = false, bool linear = false)
        {
            gameObject.transform.SetParent(anchor, false);

            _DM = DM;
            //basePanel
            Image panelImage = gameObject.AddComponent<Image>();
            panelImage.material = new Material(Shader.Find("UI/Default-Overlay"));
            panelImage.color = color ?? new Color(0, 0, 0, 0f);
            panelImage.rectTransform.sizeDelta = new Vector2(width, height);

            //Image panel
            GameObject imageObject = new GameObject("ImagePanel");
            imageObject.transform.SetParent(gameObject.transform, false);

            ImagePanel = new ImagePanel(null, null, null, null, isNormalMap, linear);
            ImagePanel.gameObject.transform.SetParent(gameObject.transform, false);
            ImagePanel.gameObject.transform.localPosition = new Vector3(panelImage.rectTransform.rect.xMax - ImagePanel.image.rectTransform.rect.xMax - 10, 0, 0);


            Transform sliderPrefab = GameObject.Instantiate<Transform>(DM.manager.configurableSliderPrefab);
            sliderPrefab.SetParent(gameObject.transform, false);
            sliderPrefab.localPosition = Vector3.zero;

            sliderDynamic = sliderPrefab.GetComponent<UIDynamicSlider>();
            sliderDynamic.transform.localPosition = new Vector2(0, -80);
            sliderDynamic.quickButtonsEnabled = false;
            sliderDynamic.rangeAdjustEnabled = false;
            sliderDynamic.defaultButtonEnabled = false;
            sliderDynamic.label = "";

            Image sliderBackground = sliderDynamic.GetComponentsInChildren<Image>().Where(x => x.name == "Panel").First();
            Color c = sliderBackground.color;
            c.a = 0.1f;
            sliderBackground.color = c;

            sliderDynamic.slider.onValueChanged.AddListener(SetAlpha);



            sliderDynamic.slider.minValue = 0;
            sliderDynamic.slider.maxValue = 1;
            sliderDynamic.slider.value = 1;


            GameObject colorObject = new GameObject("ColorPanel");
            colorObject.transform.SetParent(gameObject.transform, false);
            colorObject.transform.localScale = new Vector3(.55f, .55f, .55f);
            colorObject.transform.localPosition = new Vector2(-350, 50);


            Image colorImage = colorObject.AddComponent<Image>();
            colorImage.material = new Material(Shader.Find("UI/Default-Overlay"));
            colorImage.color = new Color(0, 1, 1, 0f);
            colorImage.rectTransform.sizeDelta = new Vector2(600, 300);
            colorImage.transform.localPosition = new Vector2(-350, 50);

            Transform colorPickerPrefab = GameObject.Instantiate<Transform>(DM.manager.configurableColorPickerPrefab);
            colorPickerPrefab.SetParent(colorImage.transform, false);
            colorPickerPrefab.localPosition = Vector3.zero;

            UIDynamicColorPicker colorPickerDynamic = colorPickerPrefab.GetComponent<UIDynamicColorPicker>();
            colorPickerDynamic.showLabel = false;
            HSVColor hsvc = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
            jcolor = new JSONStorableColor("FooColor", hsvc, ColorCallback);
            jcolor.colorPicker = colorPickerDynamic.colorPicker;


            Up = new MyButton("", new Color(0.9f , 0.6f , 0.1f));
            Up.gameObject.transform.SetParent(gameObject.transform, false);
            Up.gameObject.transform.localPosition = new Vector2(-100, 50);

            RectTransform upRT = Up.gameObject.GetComponent<RectTransform>();
            upRT.sizeDelta = new Vector2(60, 60);

            byte[] tmppng = MVR.FileManagementSecure.FileManagerSecure.ReadAllBytes(_DM.GetPackagePath() + "Custom/Scripts/Chokaphi/VAM_Decal_Maker/Icons/Up-Icon.jpg");
            Texture2D UpIcon = new Texture2D(1, 1, TextureFormat.RGBA32, false);
            UpIcon.LoadImage(tmppng);
            Up.SetIcon(UpIcon);

            Down = new MyButton("", new Color(0.1f , 0.6f, 0.8f));
            Down.gameObject.transform.SetParent(gameObject.transform, false);
            Down.gameObject.transform.localPosition = Up.gameObject.transform.localPosition + new Vector3(65,0,0);

            RectTransform dwnRT = Down.gameObject.GetComponent<RectTransform>();
            dwnRT.sizeDelta = new Vector2(60, 60);

            tmppng = MVR.FileManagementSecure.FileManagerSecure.ReadAllBytes(_DM.GetPackagePath() + "Custom/Scripts/Chokaphi/VAM_Decal_Maker/Icons/Down-Icon.jpg");
            Texture2D DownIcon = new Texture2D(1, 1, TextureFormat.RGBA32, false);
            DownIcon.LoadImage(tmppng);
            Down.SetIcon(DownIcon);

            Close = new MyButton("");
            Close.gameObject.transform.SetParent(gameObject.transform, false);
            Close.gameObject.transform.localPosition = Down.gameObject.transform.localPosition + new Vector3(160, 0, 0);

            RectTransform clsRT = Close.gameObject.GetComponent<RectTransform>();
            clsRT.sizeDelta = new Vector2(60, 60);

            tmppng = MVR.FileManagementSecure.FileManagerSecure.ReadAllBytes(_DM.GetPackagePath() + "Custom/Scripts/Chokaphi/VAM_Decal_Maker/Icons/Close-Icon.jpg");
            Texture2D CloseIcon = new Texture2D(1, 1, TextureFormat.RGBA32, false);
            CloseIcon.LoadImage(tmppng);
            Close.SetIcon(CloseIcon);

            Open = new MyButton("", new Color(.7f, .7f, .7f));
            Open.gameObject.transform.SetParent(gameObject.transform, false);
            Open.gameObject.transform.localPosition = Up.gameObject.transform.localPosition + new Vector3(-65, 0, 0);

            RectTransform opnRT = Open.gameObject.GetComponent<RectTransform>();
            opnRT.sizeDelta = new Vector2(60, 60);

            tmppng = MVR.FileManagementSecure.FileManagerSecure.ReadAllBytes(_DM.GetPackagePath() + "Custom/Scripts/Chokaphi/VAM_Decal_Maker/Icons/Open-Icon.jpg");
            Texture2D OpenIcon = new Texture2D(1, 1, TextureFormat.RGBA32, false);
            OpenIcon.LoadImage(tmppng);
            Open.SetIcon(OpenIcon);


            //Texture Path Text
            GameObject textObject = new GameObject("textPanel");
            textObject.transform.SetParent(gameObject.transform, false);
            //used to prevent text object from blocking clicks
            CanvasGroup cg = textObject.AddComponent<CanvasGroup>();
            cg.blocksRaycasts = false;

            panelText = textObject.AddComponent<Text>();    
            panelText.alignment = TextAnchor.LowerLeft;
            panelText.color = Color.black;
            panelText.font = Resources.GetBuiltinResource<Font>("Arial.ttf");
            panelText.fontSize = 30;
            panelText.fontStyle = FontStyle.Bold;
            RectTransform ptRT = panelText.GetComponent<RectTransform>();
            ptRT.sizeDelta = new Vector2(width, height);
        }

        public void SetPathText(string text)
        {
            int maxLength = 45;
            if (text.Length > maxLength)
            {
               text = text.Substring(text.Length - maxLength);
            }
            panelText.text = text;
        }

        private void SetAlpha(float value)
        {
            ImagePanel.SetAlpha(value);
        }

        private void ColorCallback(JSONStorableColor jcolor)
        {
            ImagePanel.SetColor(jcolor.colorPicker.currentColor);
        }

        public void BrowserCallBack(string path)
        {
            if (string.IsNullOrEmpty(path))
            {
                //LastDir = null;
            }
            else
            {
                SetPathText(path);
                LastDir(MVR.FileManagementSecure.FileManagerSecure.GetDirectoryName(path));
                ImagePanel.LoadPNG(path, ImagePanel);
            }
        }

        public JSONClass Save()
        {
            JSONClass decalSaveJSON = new JSONClass();
            decalSaveJSON["H"] = jcolor.val.H.ToString();
            decalSaveJSON["S"] = jcolor.val.S.ToString();
            decalSaveJSON["V"] = jcolor.val.V.ToString();
            decalSaveJSON["Alpha"] = ImagePanel.image.color.a.ToString();

            if (!string.IsNullOrEmpty(ImagePanel.Path))
            {
                string relativePath = MVR.FileManagementSecure.FileManagerSecure.NormalizePath(ImagePanel.Path);
                decalSaveJSON["Path"] = relativePath;
            }
            return decalSaveJSON;
        }

        public void CreateButtonScripts(SelectionPanel selectionPanel, ImagePanel imagePanel)
        {
            Up.button.onClick.AddListener(() =>
            {
                int currentIndex = gameObject.transform.GetSiblingIndex();
                int newIndex = currentIndex - 1;
                if (newIndex >= 0)
                {
                    imagePanel.RepositionDecalPanel(this, newIndex);
                }
            });

            Down.button.onClick.AddListener(() =>
            {
                int currentIndex = gameObject.transform.GetSiblingIndex();
                int newIndex = currentIndex + 1;
                if (newIndex < imagePanel.DecalPanels.Count)
                {
                    imagePanel.RepositionDecalPanel(this, newIndex);
                }
            });

            Close.button.onClick.AddListener(() =>
            {
                imagePanel.RemoveDecalPanel(this);
            });

            Open.button.onClick.AddListener(() =>
            {
                DecalPanel d = selectionPanel.AddDecalPanel();
                int currentIndex = gameObject.transform.GetSiblingIndex();
                int newIndex = currentIndex + 1;

                imagePanel.RepositionDecalPanel(d, newIndex);
            });
        }
    }          


    public class HeaderPanel
    {
        public GameObject gameObject = new GameObject("Header");

        public SelectionPanel ActivePanel { get; set; }

        public SelectionPanel SelectionPanelDecal { get; private set; }
        public SelectionPanel SelectionPanelSpec { get; private set; }
        public SelectionPanel SelectionPanelGloss { get; private set; }
        public SelectionPanel SelectionPanelNorm { get; private set; }
        public MyButton decalButton { get; private set; }
        public MyButton specButton { get; private set; }
        public MyButton glossButton { get; private set; }
        public MyButton normButton { get; private set; }

        private List<MyButton> buttons { get; set; }

        public Dictionary<string, SelectionPanel> selectionPanels { get; private set; }

        private Decal_Maker _DM;
        public HeaderPanel(Decal_Maker DM)
        {
            _DM = DM;
            //create a UI Spacer to Hold our Panel
            UIDynamic spacerLeft = DM.CreateSpacer(false);
            spacerLeft.height = 50;
            UIDynamic spacerRight = DM.CreateSpacer(true);
            spacerRight.height = 50;

            //Image i = spacerLeft.gameObject.AddComponent<Image>();
            //i.material = new Material(Shader.Find("UI/Default-Overlay"));
            //i.material.color = Color.green;

            gameObject.transform.SetParent(spacerLeft.transform, false);
            //center panel in UI
            gameObject.transform.localPosition += new Vector3(270, 10, 0);
            //basePanel
            Image panelImage = gameObject.AddComponent<Image>();
            panelImage.material = new Material(Shader.Find("UI/Default-Overlay"));
            panelImage.color = new Color(.5f, .5f, .5f, 1);
            panelImage.rectTransform.sizeDelta = new Vector2(1050, 70);

            GridLayoutGroup glg = gameObject.AddComponent<GridLayoutGroup>();
            glg.cellSize = new Vector2(200, 50);

            glg.constraint = GridLayoutGroup.Constraint.FixedRowCount;
            glg.constraintCount = 1;
            glg.padding = new RectOffset(10, 10, 10, 10);
            glg.spacing = new Vector2(10, 0);


            decalButton = new MyButton("Decal", new Color(1, 0.3f, 0.4f));
            decalButton.gameObject.transform.SetParent(gameObject.transform, false);
            
            specButton = new MyButton("Specular", new Color(0.95f, 0.25f, 0.91f));
            specButton.gameObject.transform.SetParent(gameObject.transform, false);
            
            glossButton = new MyButton("Gloss", new Color(0.2f, .98f, 0.2f));
            glossButton.gameObject.transform.SetParent(gameObject.transform, false);
            
            normButton = new MyButton("Normal", new Color(0.2f, 0.9f, .9f));
            normButton.gameObject.transform.SetParent(gameObject.transform, false);

            buttons = new List<MyButton>() { decalButton, specButton, glossButton, normButton };

            SelectionPanelDecal = new SelectionPanel(1050, 300, new Color(1, 1, 1, 1f), DM)
            {
                MaterialSlot = "_DecalTex",
                isNormalMap = false,
                linear = false
            };


            SelectionPanelSpec = new SelectionPanel(1050, 300, new Color(1, 1, 1, 1f), DM)
            {
                MaterialSlot = "_SpecTex",
                isNormalMap = false,
                linear = true
            };

            SelectionPanelGloss = new SelectionPanel(1050, 300, new Color(1, 1, 1, 1f), DM)
            {
                MaterialSlot = "_GlossTex",
                isNormalMap = false,
                linear = true
            };

            SelectionPanelNorm = new SelectionPanel(1050, 300, new Color(1, 1, 1, 1f), DM)
            {
                MaterialSlot = "_BumpMap",
                isNormalMap = true,
                linear = true

            };
            //now finalize panels
            SelectionPanelDecal.CreatePanels();
            SelectionPanelSpec.CreatePanels();
            SelectionPanelGloss.CreatePanels();
            SelectionPanelNorm.CreatePanels();


            SelectionPanelDecal.PanelName = "Decal";
            SelectionPanelSpec.PanelName = "Specular";
            SelectionPanelGloss.PanelName = "Gloss";        
            SelectionPanelNorm.PanelName = "Normal";

            SelectionPanelDecal.ActivateWindow(true);
            SelectionPanelSpec.ActivateWindow(false);
            SelectionPanelGloss.ActivateWindow(false);
            SelectionPanelNorm.ActivateWindow(false);

            ActivePanel = SelectionPanelDecal;

            selectionPanels = new Dictionary<string, SelectionPanel> { { "Decal", SelectionPanelDecal }, { "Specular", SelectionPanelSpec }, { "Gloss", SelectionPanelGloss }, { "Normal", SelectionPanelNorm } };
            foreach (MyButton uB in buttons)
            {
                uB.button.onClick.AddListener(() =>
                {
                    foreach (MyButton b in buttons)
                    {
                        if (b.buttonText == uB.buttonText)
                        {
                            b.Selected(true);
                            ActivePanel = selectionPanels[b.buttonText.text];
                            ActivePanel.ActivateWindow(true);
                        }
                        else
                        {
                            b.Selected(false);
                            ActivePanel = selectionPanels[b.buttonText.text];
                            ActivePanel.ActivateWindow(false);
                        }
                    }
                });
            }
        }

        public void UpdateSelectionPanels()
        {
            selectionPanels["Decal"].UpdateSelectionPanels();
            selectionPanels["Specular"].UpdateSelectionPanels();
            selectionPanels["Gloss"].UpdateSelectionPanels();
            selectionPanels["Normal"].UpdateSelectionPanels();
        }

        public void ResetTextures()
        {
            SelectionPanelDecal.ResetTextures();
            SelectionPanelSpec.ResetTextures();
            SelectionPanelGloss.ResetTextures();
            SelectionPanelNorm.ResetTextures();
        }

        public void Bake()
        {
            SelectionPanelDecal.Bake();
            SelectionPanelSpec.Bake();
            SelectionPanelGloss.Bake();
            SelectionPanelNorm.Bake();
        }
 
        public void RemoveTempDecalPanels()
        {
            SelectionPanelDecal.RemoveTempDecalPanels();
            SelectionPanelSpec.RemoveTempDecalPanels();
            SelectionPanelGloss.RemoveTempDecalPanels();
            SelectionPanelNorm.RemoveTempDecalPanels();
        }

        public void ResetTempDecalPanelTable()
        {
            SelectionPanelDecal.ResetTempDecalPanelTable();
            SelectionPanelSpec.ResetTempDecalPanelTable();
            SelectionPanelGloss.ResetTempDecalPanelTable();
            SelectionPanelNorm.ResetTempDecalPanelTable();
        }

    }

    public class MyButton
    {
        public GameObject gameObject = new GameObject("MyButton");
        public Image image { get; set; }

        public Button button { get; set; }
        public Text buttonText { get; set; }

        private DefaultControls.Resources resources = new DefaultControls.Resources();

        private Color normalColor;
        private Color unselectedColor;

        public void Selected(bool value)
        {
            if (value)
            {
                ColorBlock colors = button.colors;
                colors.normalColor = normalColor;
                button.colors = colors;
            }
            else
            {
                ColorBlock colors = button.colors;
                colors.normalColor = unselectedColor;
                button.colors = colors;
            }
        }

        public MyButton(string label = "Button", Color? color = null)
        {
            
            //transform = gameObject.transform;
            RectTransform rt = gameObject.AddComponent<RectTransform>();
            rt.sizeDelta = new Vector2(160, 30);

            GameObject textGO = new GameObject("Text");
            textGO.transform.SetParent(gameObject.transform, false);

            RectTransform tRT = textGO.AddComponent<RectTransform>();
            tRT.anchorMin = Vector2.zero;
            tRT.anchorMax = Vector2.one;
            tRT.sizeDelta = Vector2.zero;

            buttonText = textGO.AddComponent<Text>();
            buttonText.text = label;
            buttonText.alignment = TextAnchor.MiddleCenter;
            buttonText.color = Color.black;
           
            buttonText.font = Resources.GetBuiltinResource<Font>("Arial.ttf");
            buttonText.fontSize = 30;

            image = gameObject.AddComponent<Image>();
            image.material = new Material(Shader.Find("UI/Default-Overlay"));
            //image.color = color ?? Color.green;

            normalColor = color ?? new Color(0.6f,0,0,1);
            button = gameObject.AddComponent<Button>();
           
            //convert color to HSV to allow easier lighten/darken operation
            float H, S, V;
            Color.RGBToHSV(normalColor, out H, out S, out V);
            ColorBlock colors = button.colors;
            colors.highlightedColor = Color.HSVToRGB(H, S, V * 1.4f);
            colors.normalColor = normalColor;
            colors.pressedColor = normalColor;// Color.HSVToRGB(H, S, V * 1.2f);
            unselectedColor = Color.HSVToRGB(H, S, V * 0.5f);

            button.colors = colors;
        }

        public void SetIcon(Texture2D icon)
        {
            image.material.mainTexture = icon;
        }
    }

    public class SelectionPanel
    {
        public GameObject gameObject = new GameObject("Panel");

        public string PanelName { get; set; }
        public UIDynamicButton AddButton { get; private set; }
        public UIDynamicButton RemoveButton { get; private set; }

        public Dictionary<string, ImagePanel> Selections { get; private set; }
        public ImagePanel ActivePanel { get; private set; }
        public ImagePanel TorsoPanel { get; private set; }
        public ImagePanel GenitalPanel { get; private set; }
        public ImagePanel FacelPanel { get; private set; }  
        public ImagePanel LimbsPanel { get; private set; }

        public string MaterialSlot { get; set; }
        public bool isNormalMap { get; set; } = false;
        public bool linear { get; set; } = false;


        private Decal_Maker _mVR;
        private string _lastDir;
        private UIDynamic spacerLeft;
        private UIDynamic spacerRight;

        public SelectionPanel(float width, float height, Color? color, Decal_Maker mVR)
        {
            //create a UI Spacer to Hold our Panel
            spacerLeft = mVR.CreateSpacer(false);
            spacerLeft.height = 300;
            spacerRight = mVR.CreateSpacer(true);
            spacerRight.height = 300;

            //Image i = spacerLeft.gameObject.AddComponent<Image>();
            //i.material = new Material(Shader.Find("UI/Default-Overlay"));
            //i.material.color = Color.blue;

            _mVR = mVR;

             //create buttons to add/remove panels
            AddButton = mVR.CreateButton("Add Image");
            RemoveButton = mVR.CreateButton("Remove Image", true);


            gameObject.transform.SetParent(spacerLeft.transform, false);
            gameObject.transform.localPosition += new Vector3(270, 0, 0);
            //basePanel
            Image panelImage = gameObject.AddComponent<Image>();
            panelImage.material = new Material(Shader.Find("UI/Default-Overlay"));
            panelImage.color = color ?? new Color(.9f, .9f, .9f, 1);
            panelImage.rectTransform.sizeDelta = new Vector2(width, height);

            GridLayoutGroup glg = gameObject.AddComponent<GridLayoutGroup>();
            glg.cellSize = new Vector2(250, 250);
            
            glg.constraint = GridLayoutGroup.Constraint.FixedRowCount;
            glg.constraintCount = 1;
            glg.padding = new RectOffset(10,10,10,10);
            glg.spacing = new Vector2(10,0);


            AddButton.button.onClick.AddListener(() => { AddDecalPanel(); });
            RemoveButton.button.onClick.AddListener(RemoveDecalPanel);
       
        }

        public void CreatePanels()
        {
            //Torso panel
            TorsoPanel = new ImagePanel(_mVR, null, "torso", MaterialSlot, isNormalMap, linear);
            TorsoPanel.gameObject.transform.SetParent(gameObject.transform, false);
            

            //Genital panel
            GenitalPanel = new ImagePanel(_mVR, null, "genitals", MaterialSlot, isNormalMap, linear);
            GenitalPanel.gameObject.transform.SetParent(gameObject.transform, false);
            // GenitalPanel.image.transform.localPosition = TorsoPanel.gameObject.transform.localPosition + new Vector3(270, 0, 0);

            //Face panel
            FacelPanel = new ImagePanel(_mVR, null, "face", MaterialSlot, isNormalMap, linear);
            FacelPanel.gameObject.transform.SetParent(gameObject.transform, false);
            //FacelPanel.image.transform.localPosition = GenitalPanel.gameObject.transform.localPosition + new Vector3(270, 0, 0);

            //Limbs panel
            LimbsPanel = new ImagePanel(_mVR, null, "limbs", MaterialSlot, isNormalMap, linear);
            LimbsPanel.gameObject.transform.SetParent(gameObject.transform, false);
            //LimbsPanel.image.transform.localPosition = FacelPanel.gameObject.transform.localPosition + new Vector3(270, 0, 0);

            ActivePanel = TorsoPanel;
            Selections = new Dictionary<string, ImagePanel>() { { "torso", TorsoPanel }, { "genitals", GenitalPanel }, { "face", FacelPanel }, { "limbs", LimbsPanel } };
            foreach (var sel in Selections)
            {
                sel.Value.Button.onClick.AddListener(() =>
                {
                    foreach (var b in Selections)
                    {
                        if (b.Key == sel.Key)
                        {
                            b.Value.Button.image.color = Color.white;
                            b.Value.ActivateWindow(true);
                            ActivePanel = b.Value;
                        }
                        else
                        {
                            b.Value.Button.image.color = Color.grey;
                            b.Value.ActivateWindow(false);
                        }
                    }
                });
            }
            ActivateWindow(false);
        }

        public void ActivateWindow(bool value)
        {
            spacerLeft.gameObject.SetActive(value);
            spacerRight.gameObject.SetActive(value);
            AddButton.gameObject.SetActive(value);
            RemoveButton.gameObject.SetActive(value);
            ActivePanel.ActivateWindow(value);
        }

        public void UpdateSelectionPanels()
        {
            _mVR.UpdateSelectionPanels(this);
        }

        public void ResetTextures()
        {
            TorsoPanel.ResetTextures();
            GenitalPanel.ResetTextures();
            FacelPanel.ResetTextures();
            LimbsPanel.ResetTextures();
        }

        public JSONClass Save()
        {
            JSONClass jc = new JSONClass();
            foreach (var selPanel in Selections)
            {
                JSONArray Panel = new JSONArray();

                if (selPanel.Value.DecalPanels.Count > 0)
                {
                    foreach (DecalPanel d in selPanel.Value.DecalPanels)
                    {
                        JSONClass tempDecal = d.Save();
                        Panel.Add("", tempDecal);
                    }
                   
                }
                jc[selPanel.Key] = Panel;
            }
            return jc;
        }

        public void Bake()
        {
            //LogError("baking selectionPanel " + customShader.name);
            TorsoPanel.Bake();
            GenitalPanel.Bake();
            FacelPanel.Bake();
            LimbsPanel.Bake();
        }

        public DecalPanel AddDecalPanel()
        {
            DecalPanel d = AddDecalPanels(ActivePanel);
            
            return d;
        }
        
        public DecalPanel AddDecalPanel(string panelName, bool temp = false)
        {
            ImagePanel active = null;
            switch (panelName)
            {
                case "torso":
                    active = TorsoPanel;
                    break;
                case "genitals":
                    active = GenitalPanel;
                    break;
                case "face":
                    active = FacelPanel;
                    break;
                case "limbs":
                    active = LimbsPanel;
                    break;   
            }
           
            DecalPanel d = AddDecalPanels(active, temp);
    
            return d;
        }

        private DecalPanel AddDecalPanels(ImagePanel active, bool temp = false)
        {
            DecalPanel d = new DecalPanel(active.spacerLeft.transform, 1050, 300, UnityEngine.Random.ColorHSV(0f, 1f, 1f, 1f, 0.5f, 1f), _mVR, isNormalMap, linear);
            d.gameObject.transform.localPosition += new Vector3(270, 0, 0);
            active.IncreaseDecalPanel(d, temp);
            d.LastDir = SetLastDir;

            d.ImagePanel.Button.onClick.AddListener(() =>
            {
                if (_lastDir == null)
                {
                    _lastDir = "Custom/Atom/Person/Textures";
                }
                else if (_lastDir.Contains("var:"))
                {
                    _lastDir = "Custom/Atom/Person/Textures";
                }
                SuperController.singleton.GetMediaPathDialog(new FileBrowserCallback(d.BrowserCallBack), "", _lastDir);
            });

            d.CreateButtonScripts(this, active);

            return d;
        }

        public void RemoveDecalPanel()
        {
            ActivePanel.RemoveDecalPanel();   
        }

        public void RemoveTempDecalPanels()
        {
            _mVR.StartCoroutine(TorsoPanel.RemoveTempDecalPanels());
            _mVR.StartCoroutine(GenitalPanel.RemoveTempDecalPanels());
            _mVR.StartCoroutine(FacelPanel.RemoveTempDecalPanels());
            _mVR.StartCoroutine(LimbsPanel.RemoveTempDecalPanels());
        }

        public void ResetTempDecalPanelTable()
        {
            TorsoPanel.ResetTempDecalPanelTable();
            GenitalPanel.ResetTempDecalPanelTable();
            FacelPanel.ResetTempDecalPanelTable();
            LimbsPanel.ResetTempDecalPanelTable();
        }
        
        void SetLastDir(string s)
        {
            _lastDir = s;
        }

    }


    public class ImagePanel 
    {
        public GameObject gameObject = new GameObject("ImagePanel");

        public Texture Original { get; set; }
        public Image image { get; set; }
        public bool ImageLoading { get; set; }
        public string Path { get; set; }
        public Button Button { get; private set; }

        public string TextureSlot { get; set; }
        public string MaterialSlot { get; set; }
        //tell VAM image system how the texture should be loaded
        public bool isNormalMap { get; private set; }
        public bool linear { get; private set; }

        public List<DecalPanel> DecalPanels { get; set; } = new List<DecalPanel>();
        private List<DecalPanel> TempDecalPanels { get; set; } = new List<DecalPanel>();

        public UIDynamic spacerLeft;
        public UIDynamic spacerRight;

        private Decal_Maker _DM;

        public ImagePanel(Decal_Maker DM = null, Color? color = null, string textureSlot = null, string materialSlot = null, bool setNormalMap = false, bool setLinear = false )
        {
            TextureSlot = textureSlot;
            MaterialSlot = materialSlot;
            isNormalMap = setNormalMap;
            linear = setLinear;

            _DM = DM;
         
            image = gameObject.AddComponent<Image>();
            image.material = new Material(Shader.Find("UI/Default-Overlay"));
            image.color = color ?? new Color(1, 1, 1, 1f);

            RectTransform imageRect = image.rectTransform;
            imageRect.sizeDelta = new Vector2(300, 300);

            Button = gameObject.AddComponent<Button>();

            if (DM)
            {
                GameObject textObject = new GameObject("Text");
                textObject.AddComponent<RectTransform>();
                textObject.transform.SetParent(gameObject.transform, false);
                textObject.transform.localPosition = new Vector3(0, -35, 0);

                Text text = textObject.AddComponent<Text>();
                text.text = textureSlot.ToUpper() ?? "";
                text.alignment = TextAnchor.LowerCenter;
                text.color = new Color(0.196078435f, 0.196078435f, 0.196078435f, 1f);
                text.font = Resources.GetBuiltinResource<Font>("Arial.ttf");
                text.fontStyle = FontStyle.Bold;
                text.fontSize = 30;
                RectTransform component = textObject.GetComponent<RectTransform>();
                component.anchorMin = Vector2.zero;
                component.anchorMax = Vector2.one;
                component.sizeDelta = Vector2.zero;

                spacerLeft = DM.CreateSpacer(false);
                spacerLeft.height = 300;

                spacerRight = DM.CreateSpacer(true);
                spacerRight.height = 300;

                //Image i = spacerLeft.gameObject.AddComponent<Image>();
                //i.material = new Material(Shader.Find("UI/Default-Overlay"));
                //i.material.color = Color.red;

                GridLayoutGroup glg = spacerLeft.gameObject.AddComponent<GridLayoutGroup>();
                glg.cellSize = new Vector2(1050, spacerLeft.height);

                glg.constraint = GridLayoutGroup.Constraint.FixedColumnCount;
                glg.constraintCount = 1;
                //glg.padding = new RectOffset(10, 10, 10, 10);
                glg.spacing = new Vector2(0,0);

                RemoveDecalPanel();
                ActivateWindow(false);
            }
        }

        public void SetAlpha(float value)
        {
            Color newAlpha = image.color;
            newAlpha.a = value;
            image.color = newAlpha;
        }

        public void SetColor(Color color)
        {
            Color newColor = image.color;
            newColor.r = color.r;
            newColor.g = color.g;
            newColor.b = color.b;
            image.color = newColor;
        }

        public void IncreaseDecalPanel(DecalPanel d, bool temp = false)
        { 
            DecalPanels.Add(d);

            ResizeSpacers(300);

            if (temp)
            {
                TempDecalPanels.Add(d);
            }
        }

        public void RemoveDecalPanel()
        {
            if (DecalPanels.Any())
            {
                GameObject.Destroy(DecalPanels.Last().gameObject);
                DecalPanels.RemoveAt(DecalPanels.Count - 1);
            }
            ResizeSpacers(-300);
        }

 		public void RemoveDecalPanel(DecalPanel d)
        {
            bool removed = DecalPanels.Remove(d);
            if (removed)
            {
                GameObject.Destroy(d.gameObject);
                ResizeSpacers(-300);
            }
        }

        public IEnumerator RemoveTempDecalPanels()
        {
            foreach (DecalPanel d in TempDecalPanels.ToList())
            {
                yield return new WaitWhile(() => false);
                RemoveDecalPanel(d);
            }
        }

        public void ResetTempDecalPanelTable()
        {
            TempDecalPanels = new List<DecalPanel>();
        }

        public void RepositionDecalPanel(DecalPanel d, int index)
        {
           d.gameObject.transform.SetSiblingIndex(index);
           DecalPanels.Remove(d);
           DecalPanels.Insert(index, d);
        }

        public void ResetTextures()
        {
            _DM.ResetTextures(this);
        }

        public void ClearTexture()
        {
            UnityEngine.Object.Destroy(image.material);
            image.material = new Material(Shader.Find("UI/Default-Overlay"));
        }

        public void Bake()
        {
            if (DecalPanels.Count > 0)
            {
                _DM.Bake(DecalPanels, TextureSlot, MaterialSlot);
            }
        }

        private void ResizeSpacers(float value)
        {
            spacerLeft.height += value;
            spacerRight.height += value;
        }

        public void ActivateWindow(bool value)
        {
            spacerLeft.gameObject.SetActive(value);
            spacerRight.gameObject.SetActive(value);  
        }
       
        public void OnImageLoaded(ImageLoaderThreaded.QueuedImage qi)
        {
            if (qi != null)
            {
                //create new material instance so we can change texture seperatly
                image.material = new Material(Shader.Find("UI/Default-Overlay"));
                image.material.mainTexture = qi.tex;
                Path = qi.imgPath;
                 
                ImageLoading = false;
            }
        }

        //Old VAM version does not give access to load texture so use VAM image loading routine instead
        //Newer can use FileManagementSecure.FileManagerSecure or loadTexture
        //Wil call OnImageLoad once finished   
        public void LoadPNG(string filePath, ImagePanel callBack)
        {
            ImageLoaderThreaded.QueuedImage queuedImage = new ImageLoaderThreaded.QueuedImage();
            queuedImage.imgPath = filePath;
            queuedImage.forceReload = false;
            queuedImage.createMipMaps = false;
            queuedImage.isNormalMap = isNormalMap;
            queuedImage.isThumbnail = false;
            queuedImage.linear = linear;
            queuedImage.createAlphaFromGrayscale = false;
            queuedImage.compress = false;//!queuedImage.isNormalMap;

            queuedImage.callback = new ImageLoaderThreaded.ImageLoaderCallback(callBack.OnImageLoaded); // = new ImageLoaderThreaded.ImageLoaderCallback(OnImageLoaded);

            //start image load
            ImageLoaderThreaded.singleton.QueueImage(queuedImage);
            ImageLoading = true;
        }

    }

    public class PrefabCollection
    {
        public string Author { get; private set; }
        public string ItemType { get; private set; }
        public string ItemName { get; private set; }
        public string Path { get; private set; }

        public string ImagePath { get; private set; }

        public PrefabCollection(string path)
    {
            string fileName = GetFileName(path);
   
            Path = path;
            MatchCollection mc = Regex.Matches(fileName, "^(.*)_(.*)_(.*).dsgn.DecalMakerPreset.json");

            foreach (Match m in mc)
            {
                Author = m.Groups[1].Value;
                ItemType = m.Groups[2].Value;
                ItemName = m.Groups[3].Value;
                string imagePath = path.Replace(".json", ".jpg");
                if (MVR.FileManagementSecure.FileManagerSecure.FileExists(imagePath))
                {
                    ImagePath = imagePath;
                }
            }
        }


        //from C# PATH
        char[] sep = new char[] { '/', '\\' };
        private string GetFileName(string path)
        {
            int num = path.LastIndexOfAny(sep);
            if (num >= 0)
            {
                return path.Substring(num + 1);
            }
            return path;
        }
    }
    public static class TempManager
    {
        //truely temp
        static List<UnityEngine.Object> tempObjects;
        //need to keep until character or skin change
        static List<UnityEngine.Object> sessionObjects;

        static TempManager()
        {
            tempObjects = new List<UnityEngine.Object>();
            sessionObjects = new List<UnityEngine.Object>();
        }
        //Create Temp Material
        public static Material TempMaterial(Shader shader)
        {
            Material temp = new Material(shader);
            tempObjects.Add(temp);

            return temp;
        }

        //Create Temp Texture
        public static Texture2D TempTexture2D(int width, int height, TextureFormat format, bool mip = false, bool linear = false)
        {
            Texture2D temp = new Texture2D(width, height, format, mip, linear);
            tempObjects.Add(temp);
            //LogError("Stored temp objects contains " + tempObjects.Count);
            return temp;
        }
        //Create Copy of existing texture
        public static Texture2D CopyTexture(Texture2D src ,bool linear = false)
        {
            Texture2D cpy = TempTexture2D(src.width, src.height, src.format, false, linear);
            Graphics.CopyTexture(src, cpy);
           
            return cpy;
        }
        
        //copy temp into session
        public static void MakeSessionObj(UnityEngine.Object obj)
        {
            if (obj == null)
                return;

            tempObjects.Remove(obj);
            sessionObjects.Add(obj);
        }

        //remove temp texture
        public static void DestroyTexture(UnityEngine.Object obj)
        {
            if (obj == null)
                return;

            tempObjects.Remove(obj);
            UnityEngine.Object.Destroy(obj);
        }
        public static void DestoryTempObjs()
        {
            foreach (var obj in tempObjects)
            {
                UnityEngine.Object.Destroy(obj);
            }
            
            tempObjects = new List<UnityEngine.Object>();
            //LogError("Stored temp objects Destoyed");
        }

        public static void DestorySessionObjects()
        {
            foreach (var obj in sessionObjects)
            {
                UnityEngine.Object.Destroy(obj);
            }
            sessionObjects = new List<UnityEngine.Object>();
            //LogError("Stored temp objects Destoyed");
        }

        public static void DestoryAll()
        {
            foreach (var obj in tempObjects)
            {
                UnityEngine.Object.Destroy(obj);
            }
            foreach (var obj in sessionObjects)
            {
                UnityEngine.Object.Destroy(obj);
            }
            tempObjects = new List<UnityEngine.Object>();
            sessionObjects = new List<UnityEngine.Object>();
            //LogError("All temp objects Destoyed");
        }
       
    }

}
